#include <stdlib.h>
#include <errno.h>

#include "ring_buffer.h"

////////////////////////////////////////////////////////////////////////////////
struct ring_buffer * ring_buffer_new(size_t const max)
   {
   if (max < 1)
      {
      errno = EINVAL;
      return NULL;
      }

   struct ring_buffer * rb = malloc(sizeof(struct ring_buffer));
   if (NULL == rb)
      return NULL;

   rb->start = rb->count = 0;
   rb->max = max;

   rb->data = calloc(rb->max, sizeof(void *));
   if (NULL == rb->data)
      {
      free(rb);
      return NULL;
      }

   return rb;
   }

////////////////////////////////////////////////////////////////////////////////
// Returns:
// NULL on error
// elem on success and the buffer is not full
// the overwritten element on success if the buffer was full
void * ring_buffer_write(struct ring_buffer * rb, void * elem)
   {
   if ((NULL == rb) || (NULL == elem))
      {
      errno = EINVAL;
      return NULL;
      }

   size_t end = rb->start + rb->count;
   void * ret = rb->data[end % rb->max];
   rb->data[end % rb->max] = elem;
   if (rb->count == rb->max)
      rb->start = (rb->start + 1) % rb->max;
   else
      {
      rb->count++;
      ret = elem;
      }

   return ret;
   }

////////////////////////////////////////////////////////////////////////////////
void * ring_buffer_read(struct ring_buffer * rb)
   {
   if (NULL == rb)
      {
      errno = EINVAL;
      return NULL;
      }

   if (0 == rb->count)
      return NULL;

   void * t = rb->data[rb->start];
   rb->count--;
   rb->start = (rb->start + 1) % rb->max;

   return t;
   }

////////////////////////////////////////////////////////////////////////////////
size_t ring_buffer_size(struct ring_buffer * rb)
   {
   if (NULL == rb)
      {
      errno = EINVAL;
      return -1;
      }

   return rb->count;
   }

////////////////////////////////////////////////////////////////////////////////
void ring_buffer_delete(struct ring_buffer * rb)
   {
   if (NULL != rb) free(rb->data);
   free(rb);
   }
